import queue
import threading
from src.ismcore.utils.ismlogging import ism_logger

logging = ism_logger(__name__)


class ThreadQueueManager:
    def __init__(self, num_workers, processor: 'BaseProcessor'):
        self.terminated = False
        self.remainder = 0
        self.num_workers = num_workers
        self.count = 0
        self.processor = processor
        self.queue = queue.Queue()


    def start(self):
        self.workers = [threading.Thread(target=self.worker) for _ in range(self.num_workers)]
        for worker in self.workers:
            worker.daemon = True
            worker.start()

    def worker(self):
        max_wait_count = 150
        max_wait_time = 1
        wait_count = 0

        while StatusCode.RUNNING == self.processor.get_current_status():

            # we do not want to block on this,
            try:
                function = self.queue.get(timeout=max_wait_time)
            except queue.Empty:
                if wait_count >= max_wait_count:
                    total_wait_time = max_wait_time * max_wait_count
                    logging.info(f'max wait time expired, waited a total of {total_wait_time}, count: {wait_count}/{max_wait_count}')
                    self.terminated = True

                wait_count += 1
                continue

            # invoke the function
            try:
                function()
                self.remainder -= 1
                logging.info(f'completed worker task {function}, remainder: {self.remainder}')
            except Exception as e:
                logging.error(f'severe exception on worker function {e} for function: {function}')
                # raise e
            finally:
                self.queue.task_done()

    def add_to_queue(self, item):
        logging.info(f'added worker task {item} to queue at position {self.count}')
        self.count += 1
        self.remainder += 1
        self.queue.put(item)

    def wait_for_completion(self):
        self.queue.join()

    def stop_all_workers(self):
        self.terminated = True


import datetime as dt
from enum import Enum
from typing import Optional, List, Dict
from pydantic import BaseModel

class UserProfile(BaseModel):
    user_id: str
    email: Optional[str] = None
    name: Optional[str] = None
    max_agentic_units: Optional[int] = 10000    # HARD STOP - maximum compute units during a billing cycle
    created_date: Optional[dt.datetime] = dt.datetime.now(tz=dt.timezone.utc)

class UserProfileCredential(BaseModel):
    user_id: str
    type: str
    credentials: Optional[str] = None
    created_date: Optional[dt.datetime] = dt.datetime.now(tz=dt.timezone.utc)

class UserProject(BaseModel):
    project_id: Optional[str] = None
    project_name: str
    user_id: str
    created_date: Optional[dt.datetime] = dt.datetime.now(tz=dt.timezone.utc)
    updated_date: Optional[dt.datetime] = None
    deleted_date: Optional[dt.datetime] = None
    properties: Optional[str] = None


class WorkflowNode(BaseModel):
    node_id: Optional[str] = None
    node_type: str
    node_label: Optional[str] = None
    project_id: str
    object_id: Optional[str] = None
    position_x: float
    position_y: float
    width: Optional[float] = None
    height: Optional[float] = None


class WorkflowEdge(BaseModel):
    source_node_id: str
    target_node_id: str
    source_handle: str
    target_handle: str
    edge_label: str
    type: str
    animated: bool


class ProcessorStateDirection(Enum):
    INPUT = "INPUT"
    OUTPUT = "OUTPUT"


class UsageUnitType(Enum):
    TOKEN = "TOKEN"

class ProcessorStatusCode(Enum):
    # Represents that an entity or task has been created or initialized,
    # but no further action has been taken yet.
    CREATED = "CREATED"

    # Indicates that the entity or task is in the process of being routed
    # or directed to the appropriate destination or handler.
    ROUTE = "ROUTE"

    # Implies that the routing process has been completed, and the entity or
    # task has been successfully directed to the intended destination or handler.
    ROUTED = "ROUTED"

    # Suggests that the entity or task has been placed in a queue,
    # waiting to be processed or executed.
    QUEUED = "QUEUED"

    # Indicates that the entity or task is currently
    # being executed or processed.
    RUNNING = "RUNNING"

    # Implies that the execution or processing of the entity or task
    # is being forcefully terminated
    TERMINATE = "TERMINATE"

    # Suggests that the execution or processing of the entity
    # or task has been intentionally stopped or paused.
    STOPPED = "STOPPED"

    # Indicates that the entity or task has been successfully
    # executed or processed to completion.
    COMPLETED = "COMPLETED"

    # Suggests that an error or failure occurred during the execution
    # or processing of the entity or task, preventing it from being completed successfully.
    FAILED = "FAILED"


class UsageUnit(BaseModel):
    id: Optional[int] = None     # serial id
    transaction_time: Optional[dt.datetime] = None
    project_id: str
    unit_type: UsageUnitType = UsageUnitType.TOKEN
    unit_count: int
    unit_cost: float
    unit_total: float
    reference_id: Optional[str] = None
    reference_label: Optional[str] = None

    # id serial not null primary key,
    # transaction_time timestamp,
    # unit_type usage_unit_type not null,
    # unit_count int not null default 0,
    # unit_cost float null default 0,
    # unit_total  float null default 0,
    # reference_id varchar(36),
    # reference_label varchar(4000)

class InstructionTemplate(BaseModel):
    template_id: Optional[str] = None
    template_path: str
    template_content: str
    template_type: str
    project_id: Optional[str] = None


class ProcessorProvider(BaseModel):
    id: Optional[str] = None
    name: str
    version: str
    class_name: str
    user_id: Optional[str] = None
    project_id: Optional[str] = None
    created_date: Optional[dt.datetime] = dt.datetime.now(tz=dt.timezone.utc)
    updated_date: Optional[dt.datetime] = None
    routing: Optional[str] = None  # JSON string for routing configuration


class ProcessorProperty(BaseModel):
    processor_id: str
    name: str
    value: Optional[str] = None


class ProcessorPropertiesBase(BaseModel):
    requestDelay: Optional[int] = 0


class ProcessorPropertiesLM(ProcessorPropertiesBase):
    topK: Optional[int] = 0
    topP: Optional[float] = 1.0
    maxTokens: Optional[int] = 2048
    temperature: Optional[float] = 0.7
    repeatPenalty: Optional[float] = 1.0
    presencePenalty: Optional[float] = 0.0
    frequencyPenalty: Optional[float] = 0.0


class Processor(BaseModel):
    id: Optional[str] = None
    name: Optional[str] = None
    provider_id: Optional[str] = None
    project_id: str
    status: ProcessorStatusCode = ProcessorStatusCode.CREATED
    properties: Optional[dict] = None  # Changed from str to dict to match jsonb column


class ProcessorState(BaseModel):
    internal_id: Optional[int] = None
    id: Optional[str] = None
    processor_id: str
    state_id: str
    direction: ProcessorStateDirection = ProcessorStateDirection.INPUT
    status: ProcessorStatusCode = ProcessorStatusCode.CREATED

    # this does not need to be set, it is mainly used for processing input states
    # the current index is set to the highest index that was completed, only sequence +1,
    # the maximum index is set to the highest index that was completed, irrespective of sequence.
    count: Optional[int] = None
    current_index: Optional[int] = None
    maximum_index: Optional[int] = None

    def dict(self, *args, **kwargs):
        """Override dict method to exclude 'internal_id'"""
        kwargs["exclude"] = kwargs.get("exclude", set()) | {"internal_id"}
        return super().dict(*args, **kwargs)

    def json(self, *args, **kwargs):
        """Override json method to exclude 'internal_id'"""
        kwargs["exclude"] = kwargs.get("exclude", set()) | {"internal_id"}
        return super().json(*args, **kwargs)
    #
    # @property
    # def internal_id(self):
    #     return self._internal_id
    #
    # @internal_id.setter
    # def internal_id(self, internal_id):
    #     self._internal_id = internal_id


class MonitorLogEvent(BaseModel):
    log_id: Optional[int] = None
    log_type: str
    log_time: Optional[dt.datetime] = None
    internal_reference_id: Optional[int] = None
    user_id: Optional[str] = None
    project_id: Optional[str] = None
    exception: Optional[str] = None
    data: Optional[str] = None


class UnitType(Enum):
    # You'll need to define the possible values for UsageType here
    # For example:
    TOKEN = "TOKEN"
    COMPUTE = "COMPUTE"
    STORAGE = "STORAGE"


class UnitSubType(Enum):
    # You'll need to define the possible values for UsageType here
    # For example:
    INPUT = "INPUT"
    OUTPUT = "OUTPUT"
    # Add other types as needed


class Usage(BaseModel):
    id: Optional[int] = None
    transaction_time: Optional[dt.datetime] = dt.datetime.now()

    project_id: Optional[str] = None

    resource_id: str
    resource_type: str

    unit_type: UnitType
    unit_subtype: UnitSubType
    unit_count: int

    metadata: Optional[str] = None

    class Config:
        from_attributes = True


class UsageReport(BaseModel):
    user_id: str
    resource_id: Optional[str] = None
    resource_type: Optional[str] = None
    project_id: Optional[str] = None
    day: Optional[int] = None
    month: Optional[int] = None
    year: Optional[int] = None
    unit_type: Optional[str] = None
    unit_subtype: Optional[str] = None
    total: int
    total_cost: Optional[float] = 0.0


class UsageReportInstant(BaseModel):
    user_id: Optional[str] = None
    project_id: Optional[str] = None
    total: Optional[float] = 0.0
    total_cost: Optional[float] = 0.0


# Enum for user_session_access_level
class UserSessionAccessLevel(Enum):
    DEFAULT = 'default'
    ADMIN = 'admin'


# BaseModel for the 'session' table
class Session(BaseModel):
    session_id: str
    created_date: dt.datetime
    owner_user_id: str


class StateActionDefinition(BaseModel):
    id: Optional[str] = None
    state_id: str
    action_type: str
    field: Optional[str] = None
    field_options: Optional[Dict] = None
    remote_url: Optional[bool] = False
    created_date: Optional[dt.datetime] = dt.datetime.utcnow()


# BaseModel for the 'session_message' table
class SessionMessage(BaseModel):
    message_id: Optional[int] = None
    session_id: str
    user_id: str        # originated by
    original_content: Optional[str] = None
    executed_content: Optional[str] = None
    message_date: dt.datetime


# BaseModel for the 'user_session_access' table
class UserSessionAccess(BaseModel):
    user_id: str
    session_id: str
    access_level: UserSessionAccessLevel = UserSessionAccessLevel.DEFAULT
    access_date: dt.datetime


class ProcessorStateDetail(ProcessorState, Processor):
    pass



import datetime as dt
from enum import Enum
from typing import Optional

from pydantic import Json, BaseModel


class ConfigMapType(Enum):
    SECRET = "secret"
    CONFIG_MAP = "config_map"


class VaultType(Enum):
    VAULT = "Vault"
    LOCAL = "Local"
    KMS = "kms"


class Vault(BaseModel):
    id: Optional[str] = None
    name: str
    owner: Optional[str] = None
    # TODO credentials such as api keys or other if we want to support multiple vaults on a per owner basis
    type: VaultType = VaultType.LOCAL  # e.g., 'kms', 'vault', 'local', other provider
    metadata: Optional[Json] = None  # Additional metadata in JSON format
    created_at: Optional[dt.datetime] = None  # ISO timestamp
    updated_at: Optional[dt.datetime] = None  # ISO timestamp


class ConfigMap(BaseModel):
    id: Optional[str] = None
    name: str
    type: ConfigMapType
    data: dict  # JSON data stored in the configuration
    vault_key_id: Optional[str] = None  # Encryption key ID
    vault_id: Optional[str] = None  # Reference to Vault
    owner_id: Optional[str] = None  # Multi-tenancy support
    created_at: Optional[dt.datetime] = None  # ISO timestamp
    updated_at: Optional[dt.datetime] = None  # ISO timestamp
    deleted_at: Optional[dt.datetime] = None  # Optional for soft delete



import random

from ismcore.utils.general_utils import calculate_uuid_based_from_string_with_sha256_seed


def test_load_template_relative_paths():
    random.seed(128)
    random_string = "".join([chr(random.randint(1, 256)) for x in range(500)])
    genarated_uuid = calculate_uuid_based_from_string_with_sha256_seed(random_string)
    assert len(genarated_uuid) == 36



import datetime as dt
import json

from typing import Any
from nats.aio.msg import Msg

from ismcore.messaging.base_message_route_model import BaseRoute
from ismcore.model.base_model import Usage, UnitType, UnitSubType, ProcessorStatusCode
from ismcore.utils.ism_logger import ism_logger

logging = ism_logger(__name__)


class MonitoredUsage:
    def __init__(self, usage_route: BaseRoute | None = None, **kwargs):
        self.usage_route = usage_route

    async def publish_usage(self, usage: Usage):
        if not self.usage_route:
            logging.error(f"no usage route set, cannot send processor usage details to downstream usage consumer")
            return

        json_dump = usage.model_dump_json()
        result = await self.usage_route.publish(json_dump)
        return result

    async def send_usage_input_tokens(self, count: int):
        # track input token count
        usage = Usage(
            resource_id=self.processor.id, resource_type=self.provider.id,
            transaction_time=dt.datetime.utcnow(), project_id=self.processor.project_id,
            unit_type=UnitType.TOKEN, unit_subtype=UnitSubType.INPUT, unit_count=count,
        )
        await self.publish_usage(usage)

    async def send_usage_output_tokens(self, count: int):
        # track output token count
        usage = Usage(
            resource_id=self.processor.id, resource_type=self.provider.id,
            transaction_time=dt.datetime.utcnow(), project_id=self.processor.project_id,
            unit_type=UnitType.TOKEN, unit_subtype=UnitSubType.OUTPUT, unit_count=count,
        )
        await self.publish_usage(usage)


class MonitoredProcessorState:

    def __init__(self, monitor_route: BaseRoute | None = None, **kwargs):
        # super().__init__(**kwargs)
        self.monitor_route = monitor_route

    async def send_processor_state_update(
            self,
            route_id: str,
            status: ProcessorStatusCode,
            data: dict = None,
            exception: Any = None):

        monitor_message = json.dumps({
            "type": "processor_state",
            "route_id": route_id,
            "status": status.name,
            "exception": str(exception) if exception else None,
            "data": data
        })

        if not self.monitor_route:
            logging.warning(f'no monitor route defined, unable to provide processor state status updates, '
                            f'here is the monitor message dump instead: {monitor_message}')
            return

        await self.monitor_route.publish(msg=monitor_message)

    async def send_processor_state_from_consumed_message(self,
                                                         consumer_message_mapping: dict,
                                                         status: ProcessorStatusCode,
                                                         exception: Any = None,
                                                         data: dict = None):

        cmm = consumer_message_mapping  # for readability short name
        if isinstance(cmm, Msg):
            try:
                data = cmm.data.decode("utf-8")
                cmm = json.loads(data)
            except:
                raise ValueError(f'unable to extract dictionary from msg {cmm}')

        route_id = cmm['route_id'] if 'route_id' in cmm else None

        await self.send_processor_state_update(
            status=status,
            route_id=route_id,
            exception=exception,
            data=data
        )

    async def pre_execute(self, consumer_message_mapping: dict, **kwargs):
        await self.send_processor_state_from_consumed_message(
            consumer_message_mapping=consumer_message_mapping,
            status=ProcessorStatusCode.QUEUED
        )

    async def intra_execute(self, consumer_message_mapping: dict, **kwargs):
        await self.send_processor_state_from_consumed_message(
            consumer_message_mapping=consumer_message_mapping,
            status=ProcessorStatusCode.RUNNING
        )

    # TODO probably should not flipflop, we can handle this flipflop in the ism-monitor consumer?
    async def post_execute(self, consumer_message_mapping: dict, **kwargs):
        await self.send_processor_state_from_consumed_message(
            consumer_message_mapping=consumer_message_mapping,
            status=ProcessorStatusCode.COMPLETED
        )

    async def fail_validate_input_message(self, consumer_message_mapping: dict, exception: Any = None):
        # failed to execute validate input message, differs from when a processor fails to execute on a query entry
        await self.send_processor_state_from_consumed_message(
            consumer_message_mapping=consumer_message_mapping,
            status=ProcessorStatusCode.FAILED,
            exception=exception
        )

    async def fail_execute_processor_state(self,
                                           route_id: str,
                                           exception: Any,
                                           data: dict = None,
                                           **kwargs):

        # failed to execute the processor, differs from when a processor fails to execute on a query entry
        await self.send_processor_state_update(
            status=ProcessorStatusCode.FAILED,
            route_id=route_id,
            exception=exception,
            data=data
        )



from src.ismcore.embedding.semantic_distance import BasicSemanticSearch

reference_sentence = 'Cows are thoughtful, intelligent beings with family structures.'
other_sentences = ['Cows are sentient intelligent beings',
                   'Cows are a food source for humans',
                   'Cows are farm animals used by humans for their milk and meat',
                   'Cows are protesting against having been domesticated farm animals.',
                   'Cows love being domesticated animals used for their meat',
                   'Cows do not want to be a food source for humans',

                   # same as reference for baseline
                   'Cows are domesticated farm animals, they are only used for their meat, milk and skin',
                   'Kühe sind domestizierte Nutztiere, die hauptsächlich wegen ihres Fleisches, ihres Leders und ihrer Milch genutzt werden.',
                   'Les vaches sont des animaux domestiques utilisés principalement pour leur viande, leur cuir et leur lait.']

semantics = [
    BasicSemanticSearch(model_name="bert-base-uncased"),
    BasicSemanticSearch(model_name="sentence-transformers/all-MiniLM-L6-v2"),
    BasicSemanticSearch(model_name="sentence-transformers/all-mpnet-base-v2")
]

for semantic in semantics:
    print(f''.join(['-' for _ in range(1, 80)]))
    print(f'Using the {semantic} with model {semantic.model_name} to calculate sentence similarity distance calculations')
    distances_from_reference = semantic.calculate_distances(reference_sentence=reference_sentence,
                                                            other_sentences=other_sentences)

    print(f"Calculated semantic distances from reference sentence: {reference_sentence}")
    for idx, distance in enumerate(distances_from_reference):
        print(f' - {distance}\t\t{other_sentences[idx]}')


#
# print(f''.join(['-' for _ in range(1, 80)]))
# print(f"Calculated BLEU, METEOR and TER scores from reference sentence: {reference_sentence}")
# for idx, sentence in enumerate(other_sentences):
#     bleu_score = calculate_bleu(reference_sentence, sentence)
#     meteor_score = calculate_meteor(reference_sentence, sentence)
#     # ter_score = calculate_ter(reference_sentence, sentence)
#
#     formatted_bleu = "{:.2e}".format(bleu_score)  # Scientific notation with 2 decimal places
#     formatted_meteor = "{:.4f}".format(meteor_score)  # Standard decimal with 4 decimal places
#     # formatted_ter = "{:.4f}".format(ter_score)  # Standard decimal with 4 decimal places
#     formatted_ter = "N/A"
#     print(f' - blue: {formatted_bleu}\t meteor: {formatted_meteor}\t ter: {formatted_ter}\t\t{other_sentences[idx]}')



import datetime

from ismcore.model.base_model import Usage, UnitType, UnitSubType


def test_usage_entity():
    usage = Usage(
        resource_id="12345",
        resource_type="/provider/selector/id",

        #
        transaction_time=datetime.datetime.utcnow(),

        #
        unit_type=UnitType.TOKEN,
        unit_subtype=UnitSubType.INPUT,
        unit_count=123,
    )

    json_data = usage.model_dump_json(indent=2)
    print(json_data)



import json
import asyncio
from typing import Any, List, Dict, Union

from ismcore.messaging.base_message_route_model import BaseRoute
from ismcore.processor.monitored_processor_state import MonitoredProcessorState
from ismcore.storage.processor_state_storage import StateMachineStorage, FieldConfig
from ismcore.utils.general_utils import build_template_text_v2
from ismcore.utils.state_utils import validate_processor_status_change
from ismcore.utils.ism_logger import ism_logger
from ismcore.model.base_model import (
    ProcessorStateDirection,
    InstructionTemplate,
    ProcessorProvider,
    ProcessorPropertiesBase,
    Processor,
    ProcessorState,
    ProcessorStatusCode)
from ismcore.model.processor_state import (
    State,
    StateDataRowColumnData,
    StateDataColumnDefinition,
    StateDataKeyDefinition,
    StateConfig,
    StateDataColumnIndex, StateConfigStream,
)

logging = ism_logger(__name__)


class StatePropagationProvider:
    async def apply_state(self, processor: 'BaseProcessor',
                          input_query_state: Any,
                          output_query_states: [dict]) -> [dict]:
        raise NotImplementedError()


class StatePropagationProviderRouter(StatePropagationProvider):

    def __init__(self, route: BaseRoute = None):
        self.route = route

    async def apply_state(self,
                          processor: 'BaseProcessor',
                          input_query_state: Any,
                          output_query_states: [dict]) -> [dict]:
        """
        Route the processed new query states from the response to a synchronization topic

        Args:
            processor (List[Dict]): The processor instance that is processing this input query state entry
            input_query_state (Any): The initial input query state.
            output_query_states (List[Dict]): The processed output query states.

        Returns:
            List[Any]: The result of applying the query states to the output state.
        """

        # create a new message for routing purposes
        route_message = {
            "route_id": processor.output_processor_state.id,
            "type": "query_state_route",
            "input_query_state": input_query_state,
            "query_state": output_query_states
        }

        await self.route.publish(json.dumps(route_message))
        return output_query_states


class StatePropagationProviderRouterStateRouter(StatePropagationProviderRouter):

    def __init__(self, route: BaseRoute, storage: StateMachineStorage):
        """
            route (BaseRoute): the route to propagate messages to, as per conditions in apply_state(..)
            storage (StateMachineStorage): The storage system used fetch a list of state id -> processors

        :param route:
        :param storage:
        """
        super().__init__(route=route)
        self.storage = storage

    async def apply_state(self,
                          processor: 'BaseProcessor',
                          input_query_state: Any,
                          output_query_states: [dict]) -> [dict]:
        """
        Persists the processed new query states from the response.

        Args:
            processor (List[Dict]): Processor instance that is processing this input query state entry
            input_query_state (Any): Initial input query state.
            output_query_states (List[Dict]): Processed output states given the input, for a processor id.

        Returns:
            List[Any]: The result of applying the query states to the output state.
        """

        output_state = processor.output_state

        # If the flag is set to false, then skip it
        if not processor.config.flag_auto_route_output_state:
            logging.debug(f'skipping auto route of output state events, for state id: {output_state.id}')
            return output_query_states

        # I know this is confusing: the current processor handles an input -> task -> output
        #
        # the `output state id` of this processor IS an `input state id' of other processor(s)
        # thus if we want the data to route from the current state to downstream states through their
        # respective processors, we need to:
        #   1. find all processors that have an `input state id` == `output state id`
        #   2. iterate each processor state route id (state -> processor)
        #   3. for each route publish current state data processed to next route hop (as per processor)
        #
        forward_routes = self.storage.fetch_processor_state_route(
            state_id=output_state.id,
            direction=ProcessorStateDirection.INPUT
        )

        # ensure there are forwarding hop(s)
        if not forward_routes:
            logging.debug(f'no forward routes found for state id: {processor.output_state.id}')
            return

        # iterate and send query states to next hops
        [await self.route.publish(json.dumps(
            {
                "type": "query_state_entry",
                "route_id": forward_route.id,
                "query_state": output_query_states,
            }
        )) for forward_route in forward_routes]


class StatePropagationProviderRouterStateSyncStore(StatePropagationProviderRouter):
    async def apply_state(self,
                          processor: 'BaseProcessor',
                          input_query_state: Any,
                          output_query_states: [dict]) -> [dict]:
        """
        Persists the processed new query states from the response.

        Args:
            processor (List[Dict]): The processor instance that is processing this input query state entry
            input_query_state (Any): The initial input query state.
            output_query_states (List[Dict]): The processed output query states.

        Returns:
            List[Any]: The result of applying the query states to the output state.
        """

        # If the flag is set and the flat is false, then skip it
        if not processor.config.flag_auto_save_output_state:
            logging.debug(f'skipping persistence of state events, for state id: {processor.output_state.id}')
            return output_query_states

        return await super().apply_state(
            processor=processor,
            input_query_state=input_query_state,
            output_query_states=output_query_states,
        )


class StatePropagationProviderCore(StatePropagationProvider):

    async def apply_state(self,
                          processor: 'BaseProcessor',
                          input_query_state: Any,
                          output_query_states: [dict]) -> [dict]:
        """
        Writes the output_query_states to the state object, in memory

        Args:
            processor (List[Dict]): The processor instance that is processing this input query state entry
            input_query_state (Any): The initial input query state.
            output_query_states (List[Dict]): The processed output query states.

        Returns:
            List[Any]: The result of applying the query states to the output state.
        """
        # Otherwise attempt to persist the data
        logging.debug(f'persisting processed new query states from response. query states: {output_query_states} ')
        return [processor.output_state.apply_query_state(  # Iterate each query state and apply it to the output state
            query_state=query_state,
            scope_variable_mappings={
                "provider": processor.provider,
                "processor": processor.processor,
                "input_query_state": input_query_state
            }
        ) for query_state in output_query_states]


class StatePropagationProviderDistributor(StatePropagationProvider):

    def __init__(self, propagators: List[StatePropagationProvider]):
        self.propagators = propagators

    async def apply_state(
            self,
            processor: 'BaseProcessor',
            input_query_state: Any,
            output_query_states: [dict]) -> [dict]:
        """
        Writes the output_query_states to the state object, in memory

        Args:
            processor (List[Dict]): The processor instance that is processing this input query state entry
            input_query_state (Any): The initial input query state.
            output_query_states (List[Dict]): The processed output query states.

        Returns:
            List[Any]: The result of applying the query states to the output state.
        """
        # iteration each propagator and invoke it
        for propagator in self.propagators:
            await propagator.apply_state(
                processor=processor,
                input_query_state=input_query_state,
                output_query_states=output_query_states
            )

        # return the final propagated output query states
        return output_query_states


class BaseProcessor(MonitoredProcessorState):

    @property
    def template(self) -> InstructionTemplate | None:
        # if not isinstance(self.config, StateConfigStream):
        #     raise ValueError("system template cannot be set for streaming configuration, use template instead")

        if self.config.template_id:
            template = self.storage.fetch_template(self.config.template_id)
            return template

        return None

    def __init__(self,
                 output_state: State,
                 state_machine_storage: StateMachineStorage,
                 provider: ProcessorProvider = None,
                 processor: Processor = None,
                 output_processor_state: ProcessorState = None,
                 state_propagation_provider: StatePropagationProvider = StatePropagationProviderCore(),
                 stream_route: BaseRoute = None,
                 **kwargs):

        super().__init__(**kwargs)

        # TODO move into a Syncable and StateRouteable feature class
        # self.sync_store_route = sync_store_route
        # self.state_router_route = state_router_route

        self.state_propagation_provider = state_propagation_provider

        self.current_status = ProcessorStatusCode.CREATED
        self.output_state = output_state
        self.storage = state_machine_storage
        self.provider = provider
        self.processor = processor
        self.output_processor_state = output_processor_state
        self.stream_route = stream_route

        logging.info(f'setting up processor: {self.processor.id if processor else None},'
                     f'provider id: {self.provider.id if provider else None}, '
                     f'provider name: {self.provider.name if provider else None}, '
                     f'provider version: {self.provider.version if provider else None}, '
                     f'config: {self.config}')

    @property
    def properties(self) -> ProcessorPropertiesBase:
        """Return typed processor properties"""
        if not self.processor.properties:
            return ProcessorPropertiesBase()
        return ProcessorPropertiesBase(**self.processor.properties)

    @property
    def config(self) -> Union[StateConfig, StateConfigStream]:
        return self.output_state.config

    @config.setter
    def config(self, config):
        self.output_state.config = config

    @property
    def data(self):
        return self.output_state.data

    @property
    def columns(self):
        return self.output_state.columns

    @columns.setter
    def columns(self, columns):
        self.output_state.columns = columns

    @property
    def mapping(self):
        return self.output_state.mapping

    def fetch_session_data(self, input_data):
        if not isinstance(input_data, dict):
            return []

        if 'session_id' not in input_data:
            return []

        user_id = input_data['source']
        session_id = input_data['session_id']
        session_history = self.storage.fetch_session_messages(
            user_id=user_id, session_id=session_id
        )

        if not session_history:
            return []

        messages_dict = [json.loads(entry.original_content) for entry in session_history]

        # messages_dict = []
        # for entry in session_history:
        #     try:
        #         entry = json.loads(entry)
        #     except:
        #         pass
        #
        #     if isinstance(entry, dict):
        #         # role = entry['role'] if 'role' in entry else 'history'
        #         # content = entry['content'].strip() if 'content' in entry else str(entry)
        #         messages_dict.append(entry)
        #     else:
        #         messages_dict.append({"role": "user", "content": str(entry)})

        return messages_dict

    def has_query_state(self, query_state_key: str, force: bool = False):
        # make sure that the state is initialized and that there is a data key
        if not self.mapping:
            return None

        # skip if not forced and state exists
        if not force and query_state_key in self.mapping:
            logging.info(f'query {query_state_key}, cached, on config: {self.config}')
            return True

        # otherwise return none, which means no state exists
        logging.info(f'query {query_state_key}, not cached, on config: {self.config}')
        return False

    def get_current_status(self):
        return self.current_status

    def update_current_status(self, new_status: ProcessorStatusCode):
        validate_processor_status_change(
            current_status=self.get_current_status(),
            new_status=new_status
        )

        self.current_status = new_status

    async def can_processor_process_data(self):
        processor = self.storage.fetch_processor(processor_id=self.processor.id)
        if not processor:
            logging.error(f'critical, unable to find processor id: {self.processor.id}, '
                          f'likely a storage implementation issue, should have not got this far.')
            return False

        # ensure that the processor status is not in terminated state
        if processor.status in [ProcessorStatusCode.TERMINATE,
                                ProcessorStatusCode.FAILED,
                                ProcessorStatusCode.STOPPED]:
            return False

        # check usage limits
        project = self.storage.fetch_user_project(project_id=processor.project_id)
        if not project:
            logging.error(f'critical, unable to find project id: {processor.project_id}, '
                          f'likely a storage implementation issue, should have not got this far.')
            return False

        user = self.storage.fetch_user_profile(user_id=project.user_id)
        usage = self.storage.fetch_usage_report(
            user_id=FieldConfig(
                field_name="user_id",
                value=project.user_id,
                use_in_group_by=True,
                use_in_where=True
            )
        )

        if usage and usage[0].total >= user.max_agentic_units:
            logging.warning(f'usage limit reached for user: {user.user_id}, '
                            f'total: {usage[0].total}, max limit: {user.max_agentic_units}')
            return False

        return True

    async def execute_set(self, input_query_state: List[dict] = None, force : bool = False):
        is_allowed_to_process = await self.can_processor_process_data()
        if not is_allowed_to_process:
            logging.debug(f'processor {self.processor.id} for {self.provider.id}' f'is in a stopped state, skipping input query processing')
            return []

        if self.config.flag_dedup_drop_enabled:
            pass  # TODO need to check input for deduplication (need to keep the hash of the input if this is enabled)

        # execute the input entry given the processor implementation
        try:
            route_id = self.output_processor_state.id

            # RUNNING: the processor is about to execute the instructions
            await self.send_processor_state_update(route_id=route_id, status=ProcessorStatusCode.RUNNING)

            # RUNNING (INTRA): the processor is executing the output instructions on the input
            output_query_states = []  # TODO not sure if we should do something with if the config is a streams?
            if self.config.flag_expect_stream:
                await self.process_input_data_stream(input_data=input_query_state)
            else:
                output_query_states = await self.process_input_data(input_data=input_query_state, force=force)

            # Apply request delay if configured
            if self.properties.requestDelay > 0:
                logging.debug(f'processor {self.processor.id} for {self.provider.id} applying request delay of {self.properties.requestDelay} ms')
                await asyncio.sleep(self.properties.requestDelay / 1000.0)  # Convert ms to seconds

            # COMPLETED: the processor has completed execution of instructions
            await self.send_processor_state_update(route_id=route_id, status=ProcessorStatusCode.COMPLETED)
            return output_query_states
        except Exception as ex:
            await self.fail_execute_processor_state(route_id=self.output_processor_state.id, exception=ex)

    async def execute_entry(self, input_query_state: dict, force: bool = False):
        """
        Executes the processor state update and processes the input data entry.

        Args:
            input_query_state (Dict): The input query state to process.
            force (bool, optional): Flag to force the process. Defaults to False.

        Returns:
            List[Dict]: The processed output query states.

        Raises:
            Exception: If an error occurs during execution.
        """
        is_allowed_to_process = await self.can_processor_process_data()
        if not is_allowed_to_process:
            logging.debug(f'processor {self.processor.id} for {self.provider.id}'
                          f'is in a stopped state, skipping input query processing')
            return []

        if self.config.flag_dedup_drop_enabled:
            pass  # TODO need to check input for deduplication (need to keep the hash of the input if this is enabled)

        # execute the input entry given the processor implementation
        try:
            route_id = self.output_processor_state.id

            # RUNNING: the processor is about to execute the instructions
            await self.send_processor_state_update(route_id=route_id, status=ProcessorStatusCode.RUNNING)

            # RUNNING (INTRA): the processor is executing the output instructions on the input
            output_query_states = []  # TODO not sure if we should do something with if the config is a streams?
            if isinstance(self.config, StateConfigStream) or self.config.flag_expect_stream:
                await self.process_input_data_stream(input_data=input_query_state)
            else:
                output_query_states = await self.process_input_data(input_data=input_query_state, force=force)

            # Apply request delay if configured
            if self.properties.requestDelay > 0:
                logging.debug(f'processor {self.processor.id} for {self.provider.id} applying request delay of {self.properties.requestDelay} ms')
                await asyncio.sleep(self.properties.requestDelay / 1000.0)  # Convert ms to seconds

            # COMPLETED: the processor has completed execution of instructions
            await self.send_processor_state_update(route_id=route_id, status=ProcessorStatusCode.COMPLETED)
            return output_query_states
        except Exception as ex:
            await self.fail_execute_processor_state(
                route_id=self.output_processor_state.id,
                exception=ex,
                data=input_query_state
            )

    async def finalize_result(self, result: dict | List[dict] | str, input_data: dict | List[dict], additional_query_state: Any) -> List[Any]:
        """
        Finalizes the result by applying the result to the output state.

        Args:
            result (Any): The result of the execution.
            input_data (Any): The initial input query state.
            additional_query_state (Any): Any additional output values.

        Returns:
            List[Any]: The final applied states.
        """

        # Apply the result from the execution
        output_query_states = await self.output_state.apply_result(
            result=result,  # the result of the execution
            input_data=input_data,  # the initial input state
            additional_query_state=additional_query_state  # any additional output values
        )

        # Apply the new query states to the state propagator, if defined
        output_query_states = await self.state_propagation_provider.apply_state(
            processor=self,
            input_query_state=input_data,
            output_query_states=output_query_states
        )

        # Apply the new query state to the persistent storage class defined
        # output_query_states = await self.save_states(
        #     input_query_state=input_query_state,
        #     output_query_states=output_query_states
        # )

        # return the results
        return output_query_states

    async def process_input_data(self, input_data: dict | List[dict], force: bool = False):
        raise NotImplementedError("event processing is not supported by this processor")

    ## TODO need to clean up these methods
    async def _stream(self, input_data: Any, template: str):
        raise NotImplementedError()

    async def process_input_data_stream(self, input_data: dict | List[dict], force: bool = False):
        if not self.stream_route:
            raise ValueError(
                f"streams are not supported by provider: {self.output_processor_state.id}, "
                f"route_id {self.output_processor_state.id}")

        if not input_data:
            raise ValueError("invalid input state, cannot be empty")

        # TODO this such a terrible HACK to use a session id for a given processor state stream
        if 'session_id' in input_data:
            session_id = input_data["session_id"]
            subject = f"processor.state.{self.output_state.id}.{session_id}"
        else:
            subject = f"processor.state.{self.output_state.id}"

        name = f"{subject}".replace("-", "_")

        # begin the processing of the prompts
        logging.debug(f"entered streaming mode, state_id: {self.output_state.id}")

        # submit to the fully qualified subject, which may include a session id
        stream_route = self.stream_route.clone(
            route_config_updates={
                "subject": subject,
                "name": name
            }
        )

        # check if template attribute exists in
        if hasattr(self, 'template'):
            if not self.template:
                template = None
            else:
                template = build_template_text_v2(self.template, input_data)
        else:
            template = None


        try:
            # submit the original request to the stream, to all subscribers of the subject
            # TODO this needs to be invoked at the LM processor level, pre-stream-processing
            if 'source' in input_data:
                await stream_route.publish(input_data['source'])
                await stream_route.publish("<<>>SOURCE<<>>")

            if 'input' in input_data:
                await stream_route.publish(input_data['input'])
                await stream_route.publish("<<>>INPUT<<>>")

            # flush the stream to ensure the messages are sent to the stream server
            await stream_route.flush()

            # build a coroutine calling the concrete implementation of the stream
            stream = self._stream(input_data=input_data, template=template)

            # execute and iterate the yielded data directly into the upstream route
            async for content in stream:
                try:
                    if isinstance(content, str):
                        await stream_route.publish(content)
                        await stream_route.flush()
                    elif content is None:
                        # Log or handle the None case if necessary
                        logging.warning('Received NoneType content, skipping...')
                    else:
                        # Handle unexpected types
                        logging.warning(f'Unexpected content type: {type(content)}')
                except Exception as critical:
                    # Provide more detailed exception handling
                    logging.critical(f'Exception encountered during streaming: {critical}', exc_info=True)

            # TODO this needs to be invoked at the LM processor level, post-stream-processing
            # submit the response message to the stream.
            await stream_route.publish("<<>>ASSISTANT<<>>")
            await stream_route.flush()

            # should gracefully close the connection
            await stream_route.disconnect()

            logging.debug(f"exit streaming mode, state_id: {self.output_state.id}")
        except Exception as exception:
            # submit the response message to the stream.
            await stream_route.publish("<<>>ERROR<<>>")
            await stream_route.flush()
            await self.fail_execute_processor_state(
                # self.output_processor_state,
                route_id=self.output_processor_state.id,
                exception=exception,
                data=input_data
            )

    #
    # async def stream_input_data_entry(self, input_query_state: dict):
    #     raise NotImplementedError("stream processing is not supported by this processor")


if __name__ == '__main__':
    # build a test state
    test_state = State(
        config=StateConfig(
            name='test state 1',
            # input_path='../states/07c5ea7bfa7e9c6ffd93848a9be3c2e712a0e6ca43cc0ad12b6dd24ebd788d6f.json',
            # output_path='../../states/',
            # output_path='../dataset/examples/states/184fef148b36325a9f01eff757f0d90af535f4259c105fc612887d5fad34ce11.json',
            primary_key=[
                StateDataKeyDefinition(name='query'),
                StateDataKeyDefinition(name='context'),
            ],
            query_state_inheritance=[
                StateDataKeyDefinition(name='query', alias='input_query'),
                StateDataKeyDefinition(name='context', alias='input_context'),
            ]
        ),
        columns={
            'query': StateDataColumnDefinition(name='query'),
            'context': StateDataColumnDefinition(name='context'),
            'response': StateDataColumnDefinition(name='response'),
            'analysis_dimension': StateDataColumnDefinition(name='response'),
            'analysis_dimension_score': StateDataColumnDefinition(name='response')
        },
        data={
            'query': StateDataRowColumnData(
                values=['tell me about dogs.', 'where do cows live?', 'why do cows exist?']),
            'context': StateDataRowColumnData(values=['Education', 'Education', 'Education']),
            'response': StateDataRowColumnData(values=['dogs are pets', 'cows live on farms', 'as a food source']),
            'analysis_dimension': StateDataRowColumnData(values=['Person-Centric', 'Person-Centric', 'Person-Centric']),
            'analysis_dimension_score': StateDataRowColumnData(values=[63, 68, 20])
        },
        mapping={
            'abc': StateDataColumnIndex(key='abc', values=[0]),
            'def': StateDataColumnIndex(key='def', values=[1]),
            'ghi': StateDataColumnIndex(key='jkl', values=[2])
        }
    )

    test_state.save_state(output_path='../states/test_state.pickle')
    test_state.save_state(output_path='../states/test_state.json')

    # when adding a new row you only provide the values, it must match the same
    # number of columns and in the order of the columns that were added, otherwise
    # there will be data / column misalignment
    test_state.add_row_data(StateDataRowColumnData(values=[
        'why are we ....?',  # query
        'Education',  # context
        'cats are ....',  # response
        'Instrumentalist',  # analysis_dimension
        45,  # analysis_dimension_score
    ]))

    test_state.add_row_data(StateDataRowColumnData(values=[
        'why are cats and dogs....?',  # query
        'Education',  # context
        'cats and dogs ....',  # response
        'Person-Centric',  # analysis_dimension
        88,  # analysis_dimension_score
    ]))

    print(test_state)



# import os
# import re
# from datetime import datetime as dt
# from typing import List, Any
# import logging
#
# # from warnings import deprecated
#
# from ..processor_state import (
#     State,
#     StateConfig, \
#     StateConfigLM
# )
#
#
# # @deprecated
# def find_state_files(search_path: str,
#                      search_recursive: bool = False,
#                      state_path_match: str = None,
#                      state_name_match: str = None,
#                      state_version_match: str = None,
#                      state_provider_match: str = None,
#                      state_model_match: str = None,
#                      state_user_template_match: str = None,
#                      state_system_template_match: str = None):
#     logging.debug(f'searching for state files in path {search_path}, '
#                   f'recursive: {search_recursive}, '
#                   f'state_path_match: {state_path_match}, '
#                   f'state_name_match: {state_name_match}')
#
#     if not os.path.exists(search_path):
#         raise Exception(f'state path does not exist: {search_path}')
#
#     files = os.listdir(search_path)
#
#     if not files:
#         logging.error(f'no state files found in {search_path}')
#         return
#
#     # final result set to merge with previous state result files
#     results = {}
#
#     for nodes in files:
#         final_path = f'{search_path}/{nodes}'
#
#         if os.path.isdir(final_path):
#             if search_recursive:
#                 logging.debug(f'recursive path {final_path}')
#
#                 # find states recursively
#                 new_results = find_state_files(search_path=final_path,
#                                                search_recursive=search_recursive,
#                                                state_path_match=state_path_match,
#                                                state_name_match=state_name_match,
#                                                state_version_match=state_version_match,
#                                                state_provider_match=state_provider_match,
#                                                state_model_match=state_model_match,
#                                                state_user_template_match=state_user_template_match,
#                                                state_system_template_match=state_system_template_match)
#                 if new_results:
#                     # append the recursive results to the current call level
#                     results = {**results, **new_results}
#                 else:
#                     logging.debug(f'no state files matched in path: {final_path}')
#
#             continue
#
#         def inv_check_match(search, value):
#             # attempt to match full path before opening state file
#             logging.debug(f'searching for pattern {search}, value: {value}, file: {final_path}')
#             if search and not re.match(search, value) and search not in value:
#                 logging.debug(f'search pattern: {search} not matched as per value: {value} on file {final_path}')
#                 return True
#
#             return False
#
#         # attempt to match full path before opening state file
#         if inv_check_match(state_path_match, final_path):
#             continue
#
#         # TODO must be a better way to only read the state configuration instead of the entire file
#         state = State.load_state(final_path)
#
#         # search the file path for a pattern match
#         if inv_check_match(state_name_match, state.config.name):
#             continue
#
#         # if is instance of state config lm
#         if isinstance(state.config, StateConfigLM):
#
#             # backwards compatibility, TODO remove this at some point
#             if 'version' not in state.config.__dict__:
#                 state.config.version = "Version 0.0"
#             try:
#                 if (inv_check_match(state_user_template_match, state.config.user_template_path) or
#                         inv_check_match(state_system_template_match, state.config.system_template_path) or
#                         inv_check_match(state_model_match, state.config.model_name) or
#                         inv_check_match(state_provider_match, state.config.provider_name) or
#                         inv_check_match(state_version_match, state.config.version)):
#                     continue
#             except Exception as e:
#                 logging.error(
#                     f'unable to process file {final_path}, attempting to evaluate config match {state.config}')
#                 logging.error(e)
#                 continue
#
#         elif state_user_template_match or state_system_template_match:
#             # skip the file is the template user match or system template
#             # match is defined and the config type is not StateConfigLM
#             continue
#
#         # append as a dictionary of file name paths and state values
#         results[final_path] = state
#
#     return results
#
#
# # @deprecated
# def search_state(path: str, filter_func: Any):
#     # TODO terrible way to search the state but it will have to do for now,
#     #  basically a linear search since it has to open the entire state file
#     #  and iterate through it.
#     state = State.load_state(path)
#     query_states = [state.get_query_state_from_row_index(x) for x in range(state.count)]
#     return [query_state for query_state in query_states if filter_func(query_state)]
#
#
# # @deprecated
# def show_state_config_modification_info(
#         old_config: StateConfig,
#         new_config: StateConfig):
#     # filter and prepare new configuration dictionary
#     new_config = {
#         key.replace('new_', ''): value
#         for key, value in new_config.items()
#         if key is not None and
#            value is not None and
#            key.startswith('new_')
#     }
#
#     old_config_json = old_config.model_dump()
#
#     # build configuration string
#     config = "\n\t".join([f'old: {key}:{value} -> {new_config[key] if key in new_config else "<unchanged>"}'
#                           for key, value
#                           in old_config_json.items()])
#
#     # project the new configuration ontop of the old config
#     old_config_json = {**old_config_json, **new_config}
#     logging.info(f'config: {config}')
#
#     new_config = old_config.model_validate(old_config_json)
#
#     # updated configuration
#     return new_config
#
#
# # @deprecated
# def show_state_column_info(state: State):
#     columns = state.columns
#     column_string = ", ".join([f'{column}' for column, value in columns.items()])
#     logging.info(f'columns = [{column_string}]')
#     return column_string
#
#
# # @deprecated
# def show_state_info(results: dict):
#     if not results:
#         raise FileExistsError(
#             f'results cannot be blank or null, please provide a list of paths and states in either as a list[str] paths or a dict[path] = state values')
#
#     if isinstance(results, List):
#         raise NotImplementedError(
#             "list not implemented yet, please use a dictionary of state paths and state key value pairs")
#
#     for state_file, state in results.items():
#         # fetch file stats
#         stat = os.stat(state_file)
#
#         # build configuration string
#         configuration_string = "\n\t".join([f'{key}:{value}'
#                                             for key, value
#                                             in state.config.model_dump().items()])
#
#         if state.columns:
#             columns_string = ", ".join([f'[{key}]' for key in state.columns.keys()])
#         else:
#             columns_string = ""
#
#         print(f'config: {configuration_string}')
#         print(f'columns: {columns_string}')
#         print(f'state row count: {implicit_count_with_force_count(state)}')
#         print(f'created on: {dt.fromtimestamp(stat.st_ctime)}, '
#               f'updated on: {dt.fromtimestamp(stat.st_mtime)}, '
#               f'last access on: {dt.fromtimestamp(stat.st_atime)}')
#
